MODULE funbox
! right hand side of singular IVP/BVP 2.1a
! ----------------------------------------------------------------------
! Version 1.00, last modified 25-01-1999.
! --- Function ---
! This module provides the ODE y'(x) = f(x,y) with all necessary
! additional information like initial conditions, boundary conditions,
! Jacobian, etc.
! --- Bibliography ---
! example 2.1a was taken from 
! P.Kofler, E.Weinm"uller,
! 'Numerical Treatment of Singular Boundary and Initial Value Problems',
! Kuwait City, Kuwait 1996 (example 1, p. 16).
! ----------------------------------------------------------------------
   USE TYPES                 ! predefined data-types
   IMPLICIT NONE
   INTEGER(KIND = i4b), PARAMETER :: box_neqn = 2
!        box_neqn ... dimension of ODE
   LOGICAL, PARAMETER :: box_linear = .true.
!        box_linear ... the problem is linear
   PRIVATE
   PUBLIC :: box_neqn, box_linear, box_ivalues, box_bvalues, box_bcond,&
             box_fcn, box_dfcn, box_true, box_truelocal

   CONTAINS
! ----------------------------------------------------------------------
SUBROUTINE box_ivalues(a, b, ya)
! --- Function and Algorithm ---
! get initial values of the ODE
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a, b, ya(:)
!        a, b ... interval to perform integration
!        ya ... initial vector
! --- Code ---
   a = 0.0_dp
   b = 1.0_dp
   ya = (/ 0.0_dp, 0.0_dp /)
   RETURN
END SUBROUTINE box_ivalues

SUBROUTINE box_bvalues(a, b, ya, yb, va, vb)
! --- Function and Algorithm ---
! get known boundary values of the ODE
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a, b, ya(:), yb(:)
!        a, b ... interval to perform integration
!        ya, yb ... initial and boundary vector of solution
   LOGICAL, INTENT(out) :: va(:), vb(:)
!        va, vb ... the ith component of ya,yb is known
! --- Code ---
   a = 0.0_dp
   b = 1.0_dp
   ya = (/ 0.0_dp, 0.0_dp /) ! exact: 0, 0
   va = (/ .false., .true. /)
   yb = (/ 1.0_dp, 0.0_dp /) ! exact: 1, 7
   vb = (/ .true., .false. /)
   RETURN
END SUBROUTINE box_bvalues

SUBROUTINE box_bcond(a, b, Ba, Bb, beta)
! --- Function and Algorithm ---
! get linear boundary conditions of the ODE: Ba*y(0)+Bb*y(1)=beta
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(out) :: a, b, Ba(:,:), Bb(:,:), beta(:)
!        a, b ... interval to perform integration
!        Ba, Bb, beta ... linear BCs
! --- Code ---
   a = 0.0_dp
   b = 1.0_dp
   ! form linear BC as Ba*y(a)+Bb*y(b)=beta
   Ba = 0.0_dp
   Ba(2,2) = 1.0_dp
   Bb = 0.0_dp
   Bb(1,1) = 1.0_dp
   beta = 0.0_dp
   beta(1) = 1.0_dp
   RETURN
END SUBROUTINE box_bcond

FUNCTION box_fcn(t, y)
! --- Function and Algorithm ---
! right hand side of ODE: f(t,y)
! --- Arguments ---
   IMPLICIT NONE
   REAL (KIND = dp), INTENT(in) :: t, y(:)
!        t ... variable
!        y ... actial value of y(t)
   REAL (KIND = dp) :: box_fcn(SIZE(y))
! --- Code ---
   ! do not evaluate at t==0!
   box_fcn = (/ y(2)/t, -1.0_dp*y(2)/t + 56.0_dp*t**6 /)
   RETURN
END FUNCTION box_fcn

FUNCTION box_dfcn(t, y)
! --- Function and Algorithm ---
! Jacobian of right hand side: f_y(t,y)
! --- Arguments ---
   IMPLICIT NONE
   REAL (KIND = dp), INTENT(in) :: t, y(:)
!        t ... variable
!        y ... actial value of y(t)
   REAL (KIND = dp) :: box_dfcn(SIZE(y),SIZE(y))
! --- Code ---
   ! do not evaluate at t==0!
   box_dfcn(1,:) = (/ 0.0_dp, 1.0_dp/t /)
   box_dfcn(2,:) = (/ 0.0_dp, -1.0_dp/t /)
   RETURN
END FUNCTION box_dfcn

FUNCTION box_true(t,y)
! --- Function and Algorithm ---
! Solution of ODE
! --- Arguments ---
   IMPLICIT NONE
   REAL (KIND = dp), INTENT(in) :: t, y(:)
!        t ... variable
!        y ... actial value of y(t), dummy just for size
   REAL (KIND = dp) :: box_true(SIZE(y))
! --- Code ---
   box_true = (/ t**7, 7.0_dp*t**7 /)
   RETURN
END FUNCTION box_true

FUNCTION box_truelocal(t, a, ya)
! --- Function and Algorithm ---
! True local solution at t if we started at t==a with y==ya. This
! solution is only available for constant coefficient matrix problems
! and depends on the eigenvalues. This case is for -1, 0.
! --- Arguments ---
   IMPLICIT NONE
   REAL (KIND = dp), INTENT(in) :: t, a, ya(:)
!        t ... variable
!        a ... start of local solution
!        ya ... value of y(a)
   REAL (KIND = dp) :: box_truelocal(SIZE(ya))
   REAL (KIND = dp) :: c1, c2, v(SIZE(ya))
! --- Code ---
   v = box_true(a,ya)
   IF (a == 0.0_dp) THEN
      c2 = 0.0_dp
      c1 = ya(1) - v(1)
   ELSE
      c2 = a*(v(2) - ya(2))
      c1 = ya(1) - v(1) - v(2) + ya(2)
   END IF
   box_truelocal = (/ c1 + c2/t, -c2/t /) + box_true(t,ya)
   RETURN
END FUNCTION box_truelocal

END MODULE funbox
