MODULE ieulidcs
! Implicit EULer IDeC method adapted for single Shooting
! ----------------------------------------------------------------------
! Version 1.04, last modified 18-06-1999.
! --- Function ---
! Integrate the IVP of ODE y'(x) = f(x,y) with constant step size using
! implicit Euler's method with global error estimation (Zadunaisky's
! technique) and Iterated Defect Correction (IDeC). Check the global
! error and readjust the step size to satisfy given global tolerance.
! --- Author ---
! Peter W. KOFLER,
! Department of Applied Mathematics and Numerical Analysis,
! Vienna University of Technology, Vienna, Austria.
! --- Bibliography ---
! M.Metcalf, J.Reid,
! 'Fortran 90 Explained', Oxford University Press (1990),
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996).
! P.E.Zadunaisky,
! 'On the Estimation of Errors Propagated in the Numerical Integration
! of Ordinary Differential Equations', Numer.Math. 27(1976), pp. 21--39.
! ----------------------------------------------------------------------
   USE TYPES                 ! predefined data-types
   IMPLICIT NONE
   REAL(KIND = dp), PARAMETER :: safety_factor_ieidec = 0.9_dp, &
                                 dumping_factor_newton = 1.0_dp
!        safety_factor_ ... safety factor for new step sizes
!        dumping_factor_ ... factor of first dumped Newton iteration
   INTEGER(KIND = i4b), PARAMETER :: max_newton_steps = 25, &
                                     max_lambda_steps = 10
!        max_newton_steps ... max. of Newton-iterations
!        max_lambda_steps ... max. of damping steps in Newton-method
   LOGICAL :: log_file = .false. ! data-output to files
   LOGICAL :: bdebug = .true. ! debug switch to check errors and to
                             ! disable step size control
   PRIVATE
   PUBLIC :: bdebug, log_file, ivp_ieuleridec

CONTAINS
! ----------------------------------------------------------------------
SUBROUTINE ivp_ieuleridec(fcn, a, y, b, h, h_min, global_tol, &
           Newton_tol, linear, method, pord, ifail, dfcn, &
           true, truelocal, solution, maxerr, maxerrx)
! Integrate the IVP of ODE y'(x) = f(x,y) with constant step size using
! implicit Euler's method with global error estimation (Zadunaisky's
! technique) and Iterated Defect Correction (IDeC). Check the global
! error and readjust the step size to satisfy given global tolerance.
! ----------------------------------------------------------------------
! Arguments
! ----------------------------------------------------------------------
   USE LINALG                ! we need LU decomposition
   IMPLICIT NONE
   INTERFACE
      FUNCTION fcn(t, y)     ! right side of the IVP, y'(t) = fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: fcn(SIZE(y))
      END FUNCTION fcn
      FUNCTION dfcn(t, y)    ! Jacobian of fcn(t, y)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: dfcn(SIZE(y), SIZE(y))
      END FUNCTION dfcn
      FUNCTION true(t, y)    ! true solution of fcn(t, y) for testing
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, y(:)
         REAL(KIND = dp) :: true(SIZE(y))
      END FUNCTION true
      FUNCTION truelocal(t, tstart, ystart)
         USE TYPES, ONLY : dp
         IMPLICIT NONE
         REAL(KIND = dp), INTENT(in) :: t, tstart, ystart(:)
         REAL(KIND = dp) :: truelocal(SIZE(ystart))
      END FUNCTION truelocal
   END INTERFACE
   REAL(KIND = dp), INTENT(in) :: a, b, h_min, Newton_tol
!        a, b ... interval to perform integration from a to b
!        h_min ... smallest absolute allowed step size, >= 0.0
!        Newton_tol ... tolerance of Newton-iteration for implicit Euler
   REAL(KIND = dp), INTENT(in out) :: y(:), h, global_tol
!        y ... (in) initial value y(a), (out) approximation for y(b)
!        h ... (in) starting step, (out) last used step, h<0 for tvp
!        global_tol ... (in) tolerance, (out) estimation of global error
   LOGICAL, INTENT(in) :: linear
!        linear ... the initial value problem is linear
   INTEGER(KIND = i4b), INTENT(in) :: method, pord
!        method ... number of iterations of IDeC (order of the method)
!        pord ... order of Zadunaisky-polynomials (>=method+1)
   INTEGER(KIND = i4b), INTENT(out) :: ifail
!        ifail ... error variable, set due to several errors
!                   0 = everything is OK
!                  -1 = illegal interval range
!                  -2 = illegal smallest step size
!                  -3 = illegal starting step size
!                  -4 = illegal global tolerance
!                  -5 = illegal Newton-tolerance
!                  -6 = illegal method or pord
!                   1 = could not complete integration due to h < h_min
!                   2 = Newton-iteration needed too many steps
!                   3 = global error did not decrease during IDeC
!                   4 = residual did not decrease during Newton
   REAL(KIND = dp), OPTIONAL, POINTER :: solution(:,:)
!        solution ... complete solution at last grid = (y_0,y_1,...y_N)
   REAL(KIND = dp), INTENT(out), OPTIONAL :: maxerr, maxerrx
!        maxerr ... maximum of global error (testing)
!        maxerrx ... position of maximum of global error (testing)
! ----------------------------------------------------------------------
! Local variables
! ----------------------------------------------------------------------
   INTEGER(KIND = i4b) :: j, iter, step, pi, feldi, p, idec, maxsteps
!        j ... loop counter
!        iter ... iteration counter
!        step ... step counter in integration from a to b
!        pi ... step counter for little steps for Zadunaisky-polynomials
!        feldi ... reference value for interpolation in yfield
!        p  ... actual order of approximation (1 to method)
!        idec ... counter IDeC iteration
!        maxsteps ... total number of steps in the integration
   REAL(KIND = dp) :: ya(SIZE(y)), x, h_used, global_est, ratio, &
                      old_h, old_est
!        ya ... store initial value for several integrations
!        x ... current position during integration in [a,b]
!        h_used ... used step size
!        global_est ... estimation of global error (Zadunaisky)
!        ratio ... step size-update ratio
!        old_h ... last used step size
!        old_est ... estimated error of solution of last iteration
   REAL(KIND = dp), ALLOCATABLE :: yfield(:,:),yfieldw(:,:),yfieldb(:,:)
!        yfield ... store all approximations to interpolate the solution
!        yfieldw ... approximations workspace to improve solution
!        yfieldb ... store basic approximations for IDeC
   CHARACTER(LEN = 12) :: name1 = 'logx.abs', & ! [x, global-err],
                          name2 = 'logx.grd', & ! [x, 0], grid
                          name3 = 'logx.est', & ! [x, global-est]
                          name4 = 'logx.gr2', & ! [x, global-err], grid
                          name5 = 'logx.lok', & ! [x, local-err]
                          name6 = 'logx.xhp', & ! [x, h]
                          name7 = 'logx.def'    ! [x, defect]
!        name? ... names of log_files
! ----------------------------------------------------------------------
! Called software components
! ----------------------------------------------------------------------
!        indirect ... ludcmp, lubksb
! ----------------------------------------------------------------------
! Code
! ----------------------------------------------------------------------
! check input and set up variables
   ifail = -1                ! check interval
   IF ((a < 0.0_dp) .OR. (b < 0.0_dp) ) RETURN
   IF (((h >0.0_dp).AND.(b <= a)).OR.((h <0.0_dp).AND.(a <= b))) RETURN
   ifail = -2                ! check smallest step size
   IF (h_min < 0.0_dp) RETURN
   IF (h_min > ABS(b-a)) RETURN
   ifail = -3                ! check starting step size
   IF (ABS(h) < h_min) h = SIGN(h_min,h)
   IF (ABS(h) > ABS(b-a)) h = SIGN(b-a,h)
   IF (h == 0.0_dp) RETURN
   ifail = -4                ! check global tolerance
   IF (global_tol <= 0.0_dp) RETURN
   IF (.not. linear) THEN
      ifail = -5             ! check Newton-tolerance
      IF (Newton_tol <= 0.0_dp) RETURN
      IF (Newton_tol > global_tol) RETURN ! should be at least as exact
   END IF
   ifail = -6                ! check method

! set up parameters for used method
   p = method                ! order of method
   IF (pord < p+1) RETURN    ! order of Zadunaisky's polynomials

   ifail = 0
   iter = 0
   ya = y                    ! store initial value y(a)
   old_h = h
   old_est = HUGE(1.0_dp)
   IF ((bdebug).AND.(PRESENT(maxerr))) THEN
      maxerr = 0.0_dp        ! max. of global error for debugging
      maxerrx = a
   END IF
! ----------------------------------------------------------------------
! do iteration until global error is below tolerance
! ----------------------------------------------------------------------
   DO
      iter = iter + 1

! calculate amount of necessary steps and adjust h to distribute grid
! points equally spaced on [a,b]
      maxsteps = CEILING((b-a)/h/(1+EPSILON(1.0_dp)))
                             ! use h*(1+EPSILON) to avoid 1 further step
                             ! due to round-off error in h
      h = (b-a)/maxsteps     ! hnew <= h
      h_used = h/pord        ! used step size for the little steps
!     WRITE(*,*) ' Maxsteps: ', maxsteps
!     WRITE(*,*) ' Step size: ',h

! now check step size again
      IF ((ABS(h)<h_min).OR.(ABS(h_used)<=MAX(a,b)*EPSILON(1.0_dp)))THEN
         IF (iter == 1) THEN ! we did not do anything
            ifail = -3       ! initial step size too small
         ELSE                ! iter > 1, so a solution is available
            ifail = 1        ! could not complete because of h < h_min
            global_tol = old_est ! that is what we got in last iteration
            h = old_h        ! with the last used step size
         END IF
         RETURN
      END IF

! allocate workspace
      IF (PRESENT(solution)) THEN
         IF (ASSOCIATED(solution)) DEALLOCATE(solution)
      END IF
      ALLOCATE(yfield(0:maxsteps*pord,SIZE(y)),yfieldw(0:pord,SIZE(y)),&
               yfieldb(0:maxsteps*pord, SIZE(y)))

      IF (log_file) THEN     ! open log files
         name1(4:4) = char(iter+64)
         OPEN(14, FILE = name1)
         WRITE(14, *) a, MAXVAL(ABS(ya - True(a,ya)))
         name2(4:4) = char(iter+64)
         OPEN(20, FILE = name2)
         WRITE(20, *) a, 0.0_dp
         name4(4:4) = char(iter+64)
         OPEN(21, FILE = name4)
         WRITE(21, *) a, MAXVAL(ABS(ya - True(a,ya)))
         name5(4:4) = char(iter+64)
         OPEN(22, FILE = name5)
         WRITE(22, *) a, 0.0_dp
         name6(4:4) = char(iter+64)
         OPEN(23, FILE = name6)
         WRITE(23, *) a, 0.0_dp
      END IF

! calculate basic approximation
      x = a
      step = 0
      y = ya                 ! initialise initial values
      yfield(step,:) = y     ! store first value, the initial value

      DO j = 1, maxsteps     ! loop for h-steps
         DO pi = 1, pord     ! loop for little steps
            IF (linear) THEN
               y = impl_lin_step(h_used,.false.)
            ELSE
               y = impl_nlin_step(h_used,.false.)
               IF (ifail /= 0) THEN ! stop integration due to error
                  DEALLOCATE(yfield,yfieldw,yfieldb) ! freeing data
                  RETURN
               END IF
            END IF
            x = x + h_used   ! go to next grid point
            step = step + 1
            yfield(step,:) = y ! store approximation
            IF (log_file) THEN
               WRITE(14, *) x, MAXVAL(ABS(y - True(x,y)))
               WRITE(22, *) x, MAXVAL(ABS(y - &
                              TrueLocal(x, x-h_used, yfield(step-1,:))))
            END IF
            IF ((bdebug).AND.(PRESENT(maxerr))) THEN
               IF (MAXVAL(ABS(y - True(x,y))) > maxerr) THEN
                   maxerr = MAXVAL(ABS(y - True(x,y)))
                   maxerrx = x
               END IF
            END IF
         END DO ! DO pi over little steps
         IF (log_file) THEN
            WRITE(20, *) x, 0.0_dp
            WRITE(21, *) x, MAXVAL(ABS(y - True(x,y)))
            WRITE(23, *) x, h
         END IF
      END DO ! DO j over steps
      yfieldb = yfield       ! store the basic approximation
! the basic approximation is of order 1 (method == 1 has been tested)

      IF (log_file) THEN
         CLOSE(14)
         CLOSE(20)
         CLOSE(21)
         CLOSE(22)
         CLOSE(23)
      END IF

! start IDeC iteration
      DO idec = 1, method
         iter = iter + 1

         IF (log_file) THEN     ! open log files
            name1(4:4) = char(iter+64)
            OPEN(14, FILE = name1)
            WRITE(14, *) a, MAXVAL(ABS(ya - True(a,ya)))
            name5(4:4) = char(iter+64)
            OPEN(22, FILE = name5)
            WRITE(22, *) a, 0.0_dp
            name3(4:4) = char(iter-1+64)
            OPEN(15, FILE = name3)
            WRITE(15,*) a, 0.0_dp
            name7(4:4) = char(iter-1+64)
            OPEN(16, FILE = name7)
            WRITE(16,*) a, 0.0_dp
         END IF

         IF ((bdebug).AND.(PRESENT(maxerr))) THEN
            IF (idec == method) THEN
               maxerr = 0.0_dp ! max. of global error for debugging
               maxerrx = a
            END IF
         END IF

! calculate method-th approximations for the neighbouring problem
         x = a               ! initialise initial values
         step = 0
         y = yfield(step,:)
         yfieldw(0,:) = y
         DO j = 1, maxsteps  ! loop for h-steps
            feldi = step
            DO pi = 1, pord  ! loop for little steps
               IF (linear) THEN
                  y = impl_lin_step(h_used,.true.)
               ELSE
                  y = impl_nlin_step(h_used,.true.)
                  IF (ifail /= 0) THEN ! stop integration due to error
                     DEALLOCATE(yfield,yfieldw,yfieldb) ! freeing data
                     RETURN
                  END IF
               END IF
               x = x + h_used ! go to next grid-points
               step = step + 1
               yfieldw(pi,:) = yfield(step,:) - (y-yfieldb(step,:))
               IF (log_file) THEN
                  WRITE(14, *) x, MAXVAL(ABS(yfieldw(pi,:) - True(x,y)))
                  WRITE(22, *) x, MAXVAL(ABS(yfieldw(pi,:) - &
                               TrueLocal(x, x-h_used, yfieldw(pi-1,:))))
                  WRITE(15, *) x, MAXVAL(ABS(y-yfieldb(step,:)))
               END IF
               IF ((bdebug).AND.(PRESENT(maxerr))) THEN
                  IF (MAXVAL(ABS(yfieldw(pi,:)-True(x,y)))>maxerr) THEN
                      maxerr = MAXVAL(ABS(yfieldw(pi,:) - True(x,y)))
                      maxerrx = x
                  END IF
               END IF
            END DO ! DO pi over little steps
            yfield(step-pord:step-1,:) = yfieldw(0:pord-1,:)
            yfieldw(0,:) = yfieldw(pord,:)
         END DO ! DO j over steps
         yfield(step,:) = yfieldw(pord,:)
         IF (log_file) THEN
            CLOSE(14)
            CLOSE(15)
            CLOSE(16)
            CLOSE(22)
         END IF

         global_est = MAXVAL(ABS(y-yfieldb(step,:)))
!        WRITE(*,*) ' error estimate ',idec-1,'th iteration: ', &
!                     global_est
         IF (.not. bdebug) THEN
            IF (global_est < global_tol) EXIT ! stop if accurate
         END IF
      END DO ! DO idec over iterations
      y = yfield(step,:)     ! store most accurate approximation
      IF (PRESENT(solution)) THEN
         ALLOCATE(solution(0:maxsteps*pord,SIZE(y)))
         solution = yfield
      END IF
      DEALLOCATE(yfield,yfieldw,yfieldb) ! before freeing data
! the iterated approximation is of order pord (has been tested)

! estimate global error with solution of the last step
!     WRITE(*,*) ' delta ', global_est
      IF (.not. bdebug) THEN
         IF (global_est < global_tol) THEN ! accept solution
            global_tol = global_est
            ifail = 0        ! OK, we did it
            RETURN
         ELSE                ! reject solution, calculate new step size
            old_h = h        ! store old h and error to estimate order
            IF (global_est >= old_est) THEN
               ifail = 3     ! error is not decreasing
               global_tol = global_est ! we got this with last step size
               RETURN
            END IF
            old_est = global_est

            ratio = (global_est/global_tol)**(1.0_dp/p)
            ratio = ratio/safety_factor_ieidec
            h = h/ratio      ! reduced step size by update ratio
!           WRITE(*,*) ' update ratio ', ratio
!           WRITE(*,*) ' new step size ', h
         END IF
      ELSE
         EXIT               ! for testing do just one big iteration
      END IF
   END DO ! iter

   RETURN
! ----------------------------------------------------------------------
! Local subroutines
! ----------------------------------------------------------------------
   CONTAINS

FUNCTION lagrang()
! Version 1.01, last modified 07-09-1998.
! --- Function and Algorithm ---
! Evaluate the polynomial sum_{j=0}^pord omega_j*f_j (lagrange-poly)
! interpolating yfield on [a,a+pord*h] at a given x. As x consists just
! of grid points, just return the value of yfield(step+1).
! required temporary storage: SIZE(y)
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: lagrang(SIZE(y))
! --- Global variables ---
!        step, yfield(:,:), y (size only)
! --- Code ---
   lagrang = yfield(step+1,:)
   RETURN
END FUNCTION lagrang

FUNCTION difflagr(x)
! Version 1.01, last modified 07-09-1998.
! --- Function and Algorithm ---
! Evaluate the derivative of the polynomial sum_{j=0}^pord omega_j*f_j
! (lagrange-polynomial) interpolating yfield on [a,a+pord*h] at x=k*h.
! required temporary storage: 2*SIZE(y)+7
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: difflagr(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: x
!        x ... value the evaluation is required
! --- Global variables ---
!        a, h_used, feldi, pord, yfield(:,:), y(size only)
! --- Local variables ---
   REAL(KIND = dp) :: w, w2, phs(SIZE(y))
!        u ... x-value shifted to [0,pord*h]
!        w, w2 ... auxiliary data
!        phs ... value of (lagrange-polynomial)' at x
   INTEGER(KIND = i4b) :: i, j, k, l
!        i ... index of evaluation point in yfield
!        j, k, l, u ... loop counter
! --- Code ---
   i = nint((x-a-h_used*feldi)/h_used) ! move interval to [0,...]
   phs = 0.0_dp              ! initialise sum of omega'_j*f_j

   DO j = 0, pord            ! calculate sum omega'_j*f_j(pord+1 sum.)
      w = 0.0_dp             ! initialise sum for omega'_j
      DO k = 0, pord         ! calculate sum omega'_j (pord sumands)
         IF ((j /= k).and.((i == j).or.(i == k))) THEN
            w2 = 1.0_dp      ! initialise k-th product in omega'_j
            DO l = 0, pord   ! calculate k-th product in omega'_j
               IF ((j /= l).and.(l /= k)) w2 = w2*(i-l)
            END DO
            w = w + w2       ! add components of omega'_j
         END IF
      END DO

      DO k = 0, pord         ! calculate the constant factor of omega_j
         IF (j /= k ) w = w/(j-k)
      END DO
      phs = phs + w*yfield(feldi+j,:) ! add omega'_j*f_j
   END DO
   difflagr = phs/h_used
   RETURN
END FUNCTION difflagr

FUNCTION impl_lin_step(h, zadu)
! Version 1.01, last modified 03-09-1998.
! --- Function and Algorithm ---
! Perform one implicit Euler step for linear or neighbouring problem.
! required temporary storage: SIZE(y)^2+4*SIZE(y)+4
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_lin_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, dfcn, x, y
!        for zadu indirect usage of yfield(:,:), feldi, pord, h_used
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), w(SIZE(y)), &
                      u, lag(SIZE(y))
!        Jac ... Jacobian for given x
!        w ... inhomogenous part of fcn
!        u ... auxiliary data
!        lag ... explicit evaluation of lagrang for IBM AIX XL Fortran
   INTEGER(KIND = i4b) :: indx(SIZE(y)), i
!        indx ... pivoting indices for LU-decomposition
!        i ... loop counter
! --- Called software components ---
!        direct ... ludcmp, lubksb, for zadu difflagr
! --- Code ---
! set up linear system Jac*y1=w for implicit Euler
   w = 0.0_dp
   u = x + h
   w = y + h*fcn(u,w)        ! inhomogenous component of problem
   IF (zadu) THEN
!     w = w + h*(difflagr(u) - fcn(u, lagrang(u)))
      lag = lagrang()
      lag = difflagr(u) - fcn(u, lag)
      IF (log_file) WRITE(16, *) u, MAXVAL(ABS(lag))
      w = w + h*lag
   END IF
   Jac = -h*dfcn(u,y)        ! homogenous component of problem
                             ! y dummy argument because linear
   DO i = 1, SIZE(y)         ! loop over diagonal, add identity
      Jac(i,i) = Jac(i,i) + 1.0_dp
   END DO ! DO i over diagonal

! solve linear system Jac*y1=w, destroying Jac and yielding y1 in w
   CALL ludcmp(Jac, indx)
   CALL lubksb(Jac, indx, w)

   impl_lin_step = w         ! y1 = new solution for y
   RETURN
END FUNCTION impl_lin_step

FUNCTION impl_nlin_step(h, zadu)
! Version 1.03, last modified 18-11-1998.
! --- Function and Algorithm ---
! Perform one implicit Euler step for non-linear fcn or the neighbouring
! problem using the damped Newton's method
! required temporary storage: SIZE(y)^2+8*SIZE(y)+8
! --- Bibliography ---
! Deuflhard,
! 'Numerische Mathematik: Eine algorithmisch orientierte Einfhrung',
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: impl_nlin_step(SIZE(y))
   REAL(KIND = dp), INTENT(in) :: h
!        h ... step size to perform implicit Euler step
   LOGICAL, INTENT(in) :: zadu
!        zadu ... use Zadunaisky polynomials for neighbouring problem
! --- Global variables ---
!        fcn, dfcn, x, y, Newton_tol, constant max_newton_steps
!        for zadu indirect usage of yfield(:,:), feldi, pord, h_used
! --- Local variables ---
   REAL(KIND = dp) :: Jac(SIZE(y),SIZE(y)), res(SIZE(y)), lambda, &
                      delta(SIZE(y)), y1(SIZE(y)), y2(SIZE(y)), u, &
                      lag(SIZE(y)), nrmdeltak, w(SIZE(y))
!        Jac ... Jacobian for given x
!        res ... Newton-residual
!        delta ... Newton-increment
!        nrmdeltak ... norm of Newton-increment (delta)
!        y1 ... approximation of step
!        y2 ... approximation of step used by damping
!        u,w ... auxiliary data, work
!        lag ... explicit evaluation of lagrang
!        lambda ... last damping factor
   INTEGER(KIND = i4b) :: indx(SIZE(y)), i, index, lindex
!        indx ... pivoting indices for LU-decomposition
!        i ... loop counter
!        index ... iteration count in Newton-iteration
!        lindex ... iteration count in damping loop
! --- Called software components ---
!        direct ... ludcmp, lubksb, for zadu difflagr
! --- Code ---
   y1 = y                    ! initial approximation for y1
   u = x + h                 ! target point
   lambda = dumping_factor_newton/2.0_dp ! initial damping factor

! calculate 1st Newton-residual
   res = y+h*fcn(u, y1) - y1 ! calculate Newton-residual
   IF (zadu) THEN
!     res = res + h*(difflagr(u) - fcn(u, lagrang(u)))
      lag = lagrang()
      lag = difflagr(u) - fcn(u, lag)
      res = res + h*lag
   END IF

! start Newton-iteration
   index = 0
   DO
      index = index + 1

! set up linear system Jac*y1^(i) = delta for Newton-increment
      delta = res
      Jac = h*dfcn(u, y1)    ! Jacobian
      DO i = 1, SIZE(y)      ! loop over diagonal, subtract identity
         Jac(i,i) = Jac(i,i) - 1.0_dp
      END DO ! DO i over diagonal

! solve linear system, destroying Jac and yielding delta_y1 in delta
      CALL ludcmp(Jac, indx)
      CALL lubksb(Jac, indx, delta)

! exit iteration if accurate
      nrmdeltak = MAXVAL(ABS(delta)) ! size of Newton-increment
      IF ((MAXVAL(ABS(res)) < Newton_tol) .AND. &
          (nrmdeltak < Newton_tol*(1.0_dp+MAXVAL(ABS(y1))))) EXIT

! update solution and iterate damping factors
      lambda = MIN(1.0_dp,lambda*2.0_dp)
      lindex = 0
      DO
         lindex = lindex + 1
         y2 = y1 - lambda*delta

! calculate new Newton-residual
         res = y+h*fcn(u, y2) - y2 ! calculate Newton-residual
         IF (zadu) res = res + h*lag

! check "monotonic test" of the residuals by the increments
         w = res             ! simplified Newton-increment
         CALL lubksb(Jac, indx, w)

         IF (MAXVAL(ABS(w)) <= (1.0_dp-lambda/2.0_dp)*nrmdeltak) THEN
            y1 = y2          ! "monotonic test" is true
            EXIT             ! therefore take this solution
         END IF

! shorten increment's length by 2
         lambda = lambda/2.0_dp
         IF (lindex > max_lambda_steps) THEN
            ifail = 4        ! error, residual did not decrease
            impl_nlin_step = 0.0_dp
            RETURN
         END IF
      END DO ! lambda dumping
      IF (index > max_newton_steps) THEN
         ifail = 2           ! error, too many iterations
         impl_nlin_step = 0.0_dp
         RETURN
      END IF
   END DO ! Newton-loop

   IF ((zadu) .AND. (log_file)) WRITE(16, *) u, MAXVAL(ABS(lag))

   impl_nlin_step = y1       ! iterated y1 = new solution for y
   RETURN
END FUNCTION impl_nlin_step

END SUBROUTINE ivp_ieuleridec

END MODULE ieulidcs
