PROGRAM MSHT
! Multiple SHooTing with implicit Euler IDeC method
! ----------------------------------------------------------------------
! Version 1.00, last modified 21-04-1999.
! --- Function ---
! Do multiple shooting for singular BVPs with separated BCs by using a
! black-box (module "ieuleridcs", subroutine "ivp_ieuleridec"). For the
! first mesh point reduced superposition is used. Non-linear problems
! are solved with damped Newton method. First we initialise all our
! variables, then we branch for linear or non-linear case and last we
! calculate the whole solution. The special case of single shooting is
! included.
! --- Author ---
! Peter W. KOFLER,
! Department of Applied Mathematics and Numerical Analysis,
! Vienna University of Technology, Vienna, Austria.
! ----------------------------------------------------------------------
   USE types                 ! data types double and long integer
   USE linalg                ! linear algebra for LU decomposition
   USE ieulidcs              ! implicit Euler with IDeC IVP solver as
                             ! black-box for the shooting algorithm
   USE funbox                ! user input: the BVP of ODE
   USE bigfun                ! big function for simulates integration
! ----------------------------------------------------------------------
!        Constants
! ----------------------------------------------------------------------
   IMPLICIT NONE
   INTEGER(KIND = i4b), PARAMETER :: neqn = box_neqn
!        neqn ... dimension of system of ODEs
   INTEGER(KIND = i4b), PARAMETER :: max_newton_steps = 50, &
                                     max_lambda_steps = 20
!        max_newton_steps ... maximum of steps for Newton-iteration used
!        max_lambda_steps ... max. of damping steps in Newton-method
   REAL(KIND = dp), PARAMETER :: dumping_factor_newton = 1.0_dp, &
                                 maxinitstart = 1.0D-10, &
                                 maxinitmul = 1.2_dp ! 5.0 for quickplot
!        dumping_factor_ ... factor of first dumped Newton iteration
!        maxinitstart ... start value of error in estimation of IV s
!        maxinitmul ... multiplicate maxinit by this for next loop
   CHARACTER(LEN = 12) :: name1 = 'log0.nwt',& ! [iter, |res|, |incr|]
                          name2 = 'log0.cnv',& ! [maxinit,maxit,glob.F.]
                          name4 = 'log0.ord'   ! [|res|, p, c]
   LOGICAL :: debug = .true., & ! we know the solution
              verbose = .false. ! do some output
   INTEGER(KIND = i4b), PARAMETER :: initloop = 0 ! 150
!        initloop ... number of loops duplicating maxinitial
! ----------------------------------------------------------------------
!        Local variables
! ----------------------------------------------------------------------
! ========= properties of the BVP =========
!        neqn (integer,fun) ... dimension of system (USER)
!        linear (logical,fun) ... problem is linear (USER)
   REAL(KIND = dp) :: a, b
!        a,b ... interval to perform shooting (USER)
   REAL(KIND = dp) :: Ba(neqn,neqn), Bb(neqn,neqn), beta(neqn)
!        Ba*y(a)+Bb*y(b)=beta ... linear BCs of BVP (USER)
   REAL(KIND = dp), ALLOCATABLE :: y0(:,:)
!        y0 ... exact y(t_i) at mesh points for testing
   INTEGER(KIND = i4b) :: dimq, dimp
!        dimq ... number of separated ICs at a = dimension of span(Q)
!        dimp ... complement number n-p = dimension of kernel(M)
! ========= properties of the shooting mesh =========
   INTEGER(KIND = i4b) :: nmesh, imesh, nlines
!        nmesh ... number of mesh points including a
!        imesh ... loop counter for mesh points
!        nlines ... number of total lines = dimp+(nmesh-1)*neqn
   REAL(KIND = dp), ALLOCATABLE :: mesh(:)
!        mesh ... mesh points
! ========= properties of the black-box IVP solver =========
   REAL(KIND = dp) :: basictol, tol, eps, hmin
!        basictol ... global error tolerance for IVP solver
!        tol ... global error tolerance and global error estimation
!        eps ... Newton residual tolerance (only non-linear problems)
!        hmin ... smallest allowed step size
   REAL(KIND = dp), ALLOCATABLE :: h(:), h2(:)
!        h ... step size for each mesh
!        h2 ... step size calculating Jacobian for each mesh
   INTEGER(KIND = i4b) :: ifail, maxidec, pord
!        ifail ... error flag for black-box
!        maxixdec ... number of IDeC iterations
!        pord ... order of Zadunaisky polynomial
   REAL(KIND = dp) :: maxerr, maxerrx, amaxerr, amaxerrx, yb(neqn)
!        maxerr ... maximum of global error on [t_i,t_i+1] (debug)
!        maxerrx ... position of this maxerr (debug)
!        amaxerr ... maximum of global error on [a,b] (debug)
!        amaxerrx ... position of this amaxerr (debug)
!        yb ... result inhomogenous problem/IVs for last integration
! ========= variables for linear case (reduced superposition) =========
   REAL(KIND = dp), ALLOCATABLE :: Aqr(:,:), R(:,:), Q(:,:)
!        Aqr ... matrix to QR decompose the separated part of Ba
!        Q,R ... QR decomposition of (p times n) IVs
   REAL(KIND = dp), ALLOCATABLE :: grY0(:,:,:), grY1(:,:), sbeta(:), &
                    ya(:,:), Z(:)
!        grY0 ... imesh=0, ICs homogenous nxq problem (the kernel of M)
!                 imesh>0, ICs homogenous nxn problem (fundamental sol.)
!        grY1 ... result of homogenous nxq/nxn problems at t_i
!        sbeta ... right hand side of LES Q*s=sbeta
!        ya ... estimated y_i at mesh points
!        Z ... IVs for big system of ICs
   INTEGER(KIND = i4b), ALLOCATABLE :: luindx(:)
!        luindx ... index for pivoting LU-decomposition
! ========= variables for non-linear case (Newton Iteration) =========
   REAL(KIND = dp) :: sqrteps, ya0(neqn), lambda, maxinitial, nrmF, &
                      Beps, nrmds, nrmo1ds, nrmo2ds, dsp, dsc
!        sqrteps ... delta for numerically differentiating (non-linear)
!        ya0 ... inhomogenous component of reduced superposition, 1.mesh
!        lambda ... last damping factor
!        maxinitial ... relative difference to exact initial value
!        nrmF ... norm of residual |F(s)|
!        Beps ... Newton epsilon = tolerance to residual
!        nrmds ... norm of Newton increment |ds|
!        nrmo1ds, nrmo2ds ... norm of old Newton increment |ds-1|,|ds-2|
!        dsp, dsc ... order and order constant of Newton increments
   REAL(KIND = dp), ALLOCATABLE :: s(:), F(:), deltas(:), deltaF(:,:), &
                    Za(:), Ze(:), delta(:), slambda(:), deltaslam(:)
!        s ... IVs
!        F ... residual, right hand side of LES F'(s)*ds=-F(s)
!        deltaF ... F'(s), Jakobian of Newton's method
!        deltas ... increment
!        Za ... IVs of big system calculating Jacobian
!        Ze ... EVs of big system calculating Jacobian
!        delta ... vector of differential quotients for Jacobian
!        slambda ... test value for damping
!        deltaslam ... sumplified increment
   INTEGER :: iloop, iter, lindex
!        iloop ... counter increasing relative difference to IV
!        iter ... number of Newton iteration
!        lindex ... iteration count in damping loop
! ========= misc. =========
   INTEGER(KIND = i4b) :: i, j, k, l
!        i ... loop counter
!        j,k,l ... precalculated indices

! ----------------------------------------------------------------------
!        Code
! ----------------------------------------------------------------------
   WRITE(*,*) 'Multiple Shooting (implicit Euler Method with IDeC)'

   ! initialise the BVP with linear BCs
   CALL box_bcond(a, b, Ba, Bb, beta)

   IF (box_linear) THEN
      WRITE(*,'(I3,A,F6.4,A,F6.4)') neqn,'-dimensional linear BVP on ', &
                                   a,' to ',b
   ELSE
      WRITE(*,'(I3,A,F6.4,A,F6.4)') neqn, &
                           '-dimensional non-linear BVP on ', a,' to ',b
   END IF

   ! we have Ba*y(a)+Bb*y(b)=beta unsorted, now look for separated ICs
   ! to use the method of reduced superposition
   dimq = 0
   DO i = 1, neqn
      IF (ALL(Bb(i,:) == 0.0_dp)) dimq = dimq + 1
   END DO ! i = 1, neqn
   dimp = neqn - dimq        ! complement dimension
   WRITE(*,'(A,I2,A,I2,A,I2,A)') ' dimension',neqn,' =',dimq, &
                                ' (separated) +',dimp,' (not separated)'

   ! swap lines to get these separated lines as the first dimq ones in
   ! the upper part of Ba and Bb.
   IF (dimq > 0) THEN        ! there are separated BCs, so sort them
      i = 1
      j = neqn
      DO          
        IF ((.not.ALL(Bb(i,:)==0.0_dp)).AND.(ALL(Bb(j,:)==0.0_dp))) THEN
           CALL swap(Ba(i,:),Ba(j,:)) ! swap lines of linear BC matrices
           CALL swap(Bb(i,:),Bb(j,:))
           CALL swap(beta(i),beta(j))
           i = i + 1         ! the new lines are ok
           j = j - 1
        ELSE IF (ALL(Bb(i,:) == 0.0_dp)) THEN
           i = i + 1         ! these lines are ok
        ELSE IF (.not. (ALL(Bb(i,:) == 0.0_dp))) THEN
           j = j -1
        END IF
        IF (i .ge. j) EXIT   ! all lines checked
      END DO ! sort BCs
   END IF ! (dimq > 0)

   IF (verbose) THEN
      WRITE(*,*) '**** the sorted BCs (Ba*y(a)+Bb*y(b)=beta) are ****'
      WRITE(*,*) 'Ba'
      CALL write_mat(Ba)
      WRITE(*,*) 'Bb'
      CALL write_mat(Bb)
      WRITE(*,*) 'beta'
      WRITE(*,*) beta
   END IF

! ----------------------------------------------------------------------
!        initialise the multiple shooting mesh
! ----------------------------------------------------------------------
   WRITE(*,*) 'number of mesh points in the shooting grid:'
!  READ(*,*) nmesh           ! FPR dummy read
   nmesh = 2
!  READ(*,*) hmin            ! just do one step per mesh grid
!  hmin = hmin/2.0_dp        ! start with 2 grids
!  nmesh = nint((b-a)/hmin)
   WRITE(*,*) nmesh

   nlines = dimp+(nmesh-1)*neqn
   ALLOCATE(mesh(0:nmesh), y0(neqn,0:nmesh))
   DO imesh = 0, nmesh
      mesh(imesh) = a+(b-a)*imesh/nmesh ! we use an aequidistant grid
      y0(:,imesh) = box_true(mesh(imesh),y0(:,imesh)) ! = y(t_imesh)
   END DO

! ----------------------------------------------------------------------
!        initialise the IVP Solver
! ----------------------------------------------------------------------
   log_file = .false.
   bdebug = debug            ! deactivate global error control

!        number of IDeC iterations in black-box
   maxidec = 4
   WRITE(*,'(A,A,I1)') ' number of IDeC iterations (with estimation) ',&
                                               'in black-box: ', maxidec

!        order of method and polynomial degree in black-box
   pord = maxidec+1
   WRITE(*,'(A,I1)') ' degree of Zadunaisky-polynomials: ', pord
   WRITE(*,'(A,I1)') ' order of black-box method: ', MIN(maxidec+1,pord)

!        global error tolerance
   IF (debug) THEN           ! investigate order of convergence
      WRITE(*,*) 'global error control in black-box deactivated'
      basictol = 1.0D+99     ! set high tolerance to accept any solution
   ELSE
      WRITE(*,*) 'tolerance to global error in black-box:'
      basictol = 1.0D-9
!     READ(*,*) basictol
      WRITE(*,*) basictol
   END IF

!        starting step size for each mesh grid
   WRITE(*,*) 'starting step size(H):'
   hmin = 1.0_dp
   READ(*,*) hmin            ! use same starting step size for each mesh
   ALLOCATE(h(0:nmesh-1))
   DO imesh = 0,nmesh-1
      h(imesh) = hmin*(mesh(imesh+1)-mesh(imesh)) ! h = h*(b-a)
   END DO
   hmin = 0.0_dp
   WRITE(*,*) h(0)/pord      ! real step size = h/pord

   IF (.not. box_linear) THEN ! non-linear BVP needs Newton parameters
      sqrteps = SQRT(EPSILON(1.0_dp))

      WRITE(*,*) 'Newton tolerance in black-box:'
      eps = basictol / 10.0_dp
      IF (debug) eps = 1.0D-14
!     READ(*,*) eps          ! FPR read
      WRITE(*,*) eps

      WRITE(*,*) 'Newton tolerance for shooting procedure:'
      beps = basictol        ! basically as accurate as solution
      IF (debug) beps = eps*100.0_dp
!     READ(*,*) Beps
      WRITE(*,*) Beps

     IF (debug) THEN
        WRITE(*,*) 'relative error of estimated initial value:'
        maxinitial = 0.0_dp
        IF (initloop == 0) maxinitial = 1.0D-10*5.0_dp**13
!       READ(*,*) maxinitial
        WRITE(*,*) maxinitial
     END IF
   ELSE
      eps = 0.0_dp           ! dummy parameter for black-box
      beps = 0.0_dp
      maxinitial = 0.0_dp
   END IF ! (.not. box_linear)

   IF (initloop > 0) OPEN(15, FILE = name2)
   DO iloop = 0, initloop

! ----------------------------------------------------------------------
!        calculate the IVs for reduced superposition by QR decomposition
! ----------------------------------------------------------------------
   WRITE(*,*) 'calculating initial value'
   ALLOCATE(ya(neqn,0:nmesh), grY0(neqn,neqn,0:nmesh-1))

   imesh = 0                 ! 1st mesh grid ---------------------------
   IF (dimq > 0) THEN        ! reduced superposition is possible
      ALLOCATE(R(dimq,dimq), Q(neqn,neqn), Aqr(neqn,dimq))
      ! calculate complement to separated ICs, we just need to shoot for
      ! these vectors: Phi bar = back part of Q^(-1)^T = Q
      Aqr = transpose(Ba(1:dimq,:))
      CALL qrdecomp(Aqr, Q)  ! yields Q in Q and R in Aqr
      grY0(:,1:dimp,imesh) = Q(:,dimq+1:neqn)
      R = Aqr(1:dimq,:)      ! we just need the upper part of R
      DEALLOCATE(Aqr)

      ! calculate inverse of R as we need R^(-1)^T for the special IV
      IF (dimq > 1) THEN
         WRITE(*,*) 'L1-condition of R in Q.R of Ba:', L1cond(R)
         ALLOCATE(Aqr(dimq,dimq), luindx(dimq))
         Aqr = 0.0_dp        ! eye(dimq)
         DO i = 1, dimq
            Aqr(i,i) = 1.0_dp
         END DO
         CALL ludcmp(R,luindx)
         CALL lubksbm(R,luindx,Aqr)
         ya(:,imesh) = MATMUL( MATMUL(Q(:,1:dimq), transpose(Aqr)), &
                                                           beta(1:dimq))
         DEALLOCATE(luindx, Aqr)
      ELSE
         ya(:,imesh) = Q(:,1)*beta(1)/R(1,1) ! 1-dimensional case
      END IF ! (dimq > 1)
      DEALLOCATE(Q, R)
   ELSE ! (dimq > 0)
      grY0(:,:,imesh) = 0.0_dp
      DO i = 1, neqn         ! eye(neqn) as dimp == neqn
         grY0(i,i,imesh) = 1.0_dp
      END DO
      ya(:,imesh) = 0.0_dp   ! we have to start with 0 as special IV
   END IF ! (dimq > 0)

   IF (nmesh>1) THEN         ! further mesh grids ----------------------
      DO imesh = 1, nmesh-1  ! initialise other values of mesh grid
         grY0(:,:,imesh) = 0.0_dp
         DO i = 1, neqn      ! eye(neqn)
            grY0(i,i,imesh) = 1.0_dp
         END DO
         ! other matrices than eye(neqn) used must be stored
         ya(:,imesh) = 0.0_dp ! special IV 0 in linear case
      END DO ! imesh = 1, nmesh-1
   END IF ! (nmesh>1)

   IF (verbose) THEN
      WRITE(*,*) '**** the initial values are ****'
      WRITE(*,*) 'grY0 (fundamental matrix/kernel)'
      CALL write_mat(grY0(:,1:dimp,0))
      WRITE(*,*) 'ya (special value)'
      WRITE(*,*) ya(:,0)
   END IF

   IF (box_linear) THEN      ! linear case
! ----------------------------------------------------------------------
!        multiple shooting for linear problems
! ----------------------------------------------------------------------
      ALLOCATE(sbeta(nlines), Q(nlines,nlines))
      sbeta = 0.0_dp
      Q = 0.0_dp

      ! solve for imesh==0: Y(b)=grY1 (nxq) and y(b)=yb (n)
      ! the special ICs have already been computed above
      imesh = 0              ! 1st mesh grid ---------------------------
      CALL big_init(dimp)    ! make dimensions available to other module
      ALLOCATE(grY1(neqn,dimp), Z((dimp+1)*neqn))
      DO i = 1, dimp         ! transform IVs to vector Z
         Z((i-1)*neqn+1:i*neqn) = grY0(:,i,imesh)
      END DO ! DO i over diagonal
      Z(dimp*neqn+1:(dimp+1)*neqn) = ya(:,imesh) ! Z contains IVs

      tol = basictol         ! adjust global error tolerance
      ifail = 0
      CALL ivp_ieuleridec(bigl_fcn, mesh(imesh), Z, mesh(imesh+1), &
           h(imesh), hmin, tol, eps, box_linear, maxidec, pord, ifail, &
           bigl_dfcn, dummy1, dummy2)
      IF (ifail /= 0) CALL error(ifail,.true.)
      IF (verbose) WRITE(*,*) 'used stepsize: ', h(imesh)/pord

      ! transform the results
      DO i = 1, dimp
         grY1(:,i) = Z((i-1)*neqn+1:i*neqn)
      END DO ! DO i over diagonal
      yb = Z(dimp*neqn+1:(dimp+1)*neqn)
      DEALLOCATE(Z)

      ! store values in Q of LES Q*s=sbeta
      IF (nmesh == 1) THEN    ! just single shooting as special case
         sbeta = beta(dimq+1:neqn) - MATMUL(Ba(dimq+1:neqn,:),ya(:,0))-&
                                            MATMUL(Bb(dimq+1:neqn,:),yb)
         Q = MATMUL(Ba(dimq+1:neqn,:),grY0(:,1:dimp,imesh)) + &
                                          MATMUL(Bb(dimq+1:neqn,:),grY1)
      ELSE
         sbeta(1:neqn) = yb
         Q(1:neqn,1:dimp) = -grY1
         Q((nmesh-1)*neqn+1:(nmesh-1)*neqn+dimp,1:dimp) = &
                          MATMUL(Ba(dimq+1:neqn,:),grY0(:,1:dimp,imesh))
      END IF ! (nmesh == 1)

      IF (nmesh > 1) THEN    ! further mesh grids ----------------------
        CALL big_init(neqn)  ! make dimensions available to other module
        DEALLOCATE(grY1)     ! keep only last grY1
        ALLOCATE(grY1(neqn,neqn), Z((neqn+1)*neqn))
        DO imesh = 1, nmesh-1
          j = (imesh-1)*neqn
          k = imesh*neqn
          l = (imesh+1)*neqn
          ! store IVs grY0 in Q of LES Q*s=sbeta
          Q(j+1:k,dimp+j+1:dimp+k) = grY0(:,:,imesh)

          DO i = 1, neqn     ! transform IVs to vector Z
             Z((i-1)*neqn+1:i*neqn) = grY0(:,i,imesh)
          END DO ! DO i over diagonal
          Z(neqn*neqn+1:(neqn+1)*neqn) = ya(:,imesh) ! Z contains IVs

          tol = basictol     ! adjust global error tolerance
          ifail = 0
          CALL ivp_ieuleridec(bigl_fcn, mesh(imesh), Z, mesh(imesh+1), &
               h(imesh), hmin, tol, eps, box_linear, maxidec, pord, &
               ifail, bigl_dfcn, dummy1, dummy2)
          IF (ifail /= 0) CALL error(ifail,.true.)
          IF (verbose) WRITE(*,*) 'used stepsize: ', h(imesh)/pord

          ! transform the results
          DO i = 1, neqn
             grY1(:,i) = Z((i-1)*neqn+1:i*neqn)
          END DO ! DO i over diagonal
          yb = Z(neqn*neqn+1:(neqn+1)*neqn)

          ! store values in Q of LES Q*s=sbeta
          sbeta(j+1:k) = sbeta(j+1:k) - ya(:,imesh)
          IF (imesh==nmesh-1) THEN ! last loop, use BVPs
            sbeta(k+1:k+dimp) = - MATMUL(Bb(dimq+1:neqn,:),yb) - &
                  MATMUL(Ba(dimq+1:neqn,:), ya(:,0)) + beta(dimq+1:neqn)
            Q(k+1:k+dimp,dimp+j+1:dimp+k)=MATMUL(Bb(dimq+1:neqn,:),grY1)
          ELSE
            sbeta(k+1:l) = yb
            Q(k+1:l, dimp+j+1:dimp+k) = -grY1
          END IF ! (imesh==nmesh-1)
        END DO ! imesh =1,nmesh
        DEALLOCATE(Z)
      END IF ! (nmesh > 1)

      ! calculate missing IVs "s" (in sbeta) from LES Q*s=sbeta
      ALLOCATE(luindx(nlines))
      WRITE(*,*) 'L1-norm Q: ',m1abs(Q)
      WRITE(*,*) 'L1-norm inverse(Q): ',m1absinv(Q)
      WRITE(*,*) 'L1-condition Q: ',L1cond(Q)
      CALL ludcmp(Q,luindx)
      CALL lubksb(Q,luindx,sbeta)
      DEALLOCATE(luindx)

      ! calculate solution with starting value "s" (in sbeta)
      imesh = 0              ! 1st mesh grid ---------------------------
      ya(:,imesh)=MATMUL(grY0(:,1:dimp,imesh),sbeta(1:dimp))+ya(:,imesh)

      IF (nmesh > 1) THEN    ! further mesh grids ----------------------
         DO imesh = 1,nmesh-1
            ya(:,imesh) = MATMUL(grY0(:,:,imesh), &
             sbeta(dimp+(imesh-1)*neqn+1:dimp+imesh*neqn)) + ya(:,imesh)
         END DO ! imesh = 1,nmesh-1
         ya(:,nmesh) = MATMUL(grY1, & ! known value
                 sbeta(dimp+(nmesh-2)*neqn+1:dimp+(nmesh-1)*neqn)) + yb
      ELSE
         ya(:,nmesh) = MATMUL(grY1,sbeta(1:dimp)) + yb
      END IF ! (nmesh > 1)
      DEALLOCATE(grY1,Q,sbeta,grY0)

      IF (verbose) THEN
        DO imesh = 0, nmesh
          WRITE(*,'(A,F8.6,A)') ' solution at ',mesh(imesh), &
                                   ' is                     (should be)'
          DO i = 1, neqn
            WRITE(*,*) ya(i,imesh), '                   (', &
                                                         y0(i,imesh),')'
          END DO
          WRITE(*,*) 'with global error ', &
                                    MAXVAL(ABS(ya(:,imesh)-y0(:,imesh)))
        END DO ! imesh = 0, nmesh
      END IF ! (verbose)
   ELSE
! ----------------------------------------------------------------------
!        multiple shooting for non-linear problems
! ----------------------------------------------------------------------
      ALLOCATE(h2(0:nmesh-1), s(nlines), F(nlines), deltas(nlines), &
                                                  deltaF(nlines,nlines))
      ya0 = ya(:,0)          ! store special IV, ignore other values ya
      h2 = h                 ! use separate step size for Jacobians

      ! estimate an initial value, the unknown components of y(a)
      s = 0.0_dp             ! this estimation may be very bad.

      if (debug) THEN
         imesh = 0           ! 1st mesh grid ---------------------------
         ! calculate exact s to manage amount of wrong estimation
         IF (dimq > 0) THEN  ! reduced superposition is possible
            ALLOCATE(Q(neqn,neqn), Aqr(neqn,dimq))
            Aqr = transpose(Ba(1:dimq,:))
            CALL qrdecomp(Aqr, Q) ! yields Q in Q and R in Aqr
            yb = MATMUL(transpose(Q),y0(:,imesh)-ya0)
            s(1:dimp) = yb(dimq+1:neqn)
            DEALLOCATE(Aqr,Q)
         ELSE ! dimp = neqn
            ALLOCATE(Aqr(neqn,neqn),luindx(neqn))
            Aqr = grY0(:,:,imesh)
            yb = y0(:,imesh)-ya0
            CALL ludcmp(Aqr,luindx)
            CALL lubksb(Aqr,luindx,yb)
            s(1:dimp) = yb
!           s(1:dimp) = y0(:,imesh)-ya0 ! if grY0(:,:,0) = eye(neqn)
            DEALLOCATE(luindx,Aqr)
         END IF

         IF (nmesh>1) THEN   ! further mesh grids ----------------------
            DO imesh = 1, nmesh-1
               ALLOCATE(Aqr(neqn,neqn),luindx(neqn))
               Aqr = grY0(:,:,imesh)
               yb = y0(:,imesh)
               CALL ludcmp(Aqr,luindx)
               CALL lubksb(Aqr,luindx,yb)
               s(dimp+(imesh-1)*neqn+1:dimp+imesh*neqn) = yb
               ! s = grY0(:,:,imesh)^{-1} * y0(:,imesh)
               DEALLOCATE(luindx,Aqr)
            END DO ! imesh = 1, nmesh-1
         END IF ! (nmesh>1)

         ! put an relative error of maxinitial in each coponent of s
         DO i = 1, nlines
            IF (s(i) /= 0.0_dp ) THEN
               s(i) = s(i)*(1.0_dp+maxinitial)
            ELSE
               s(i) = maxinitial
            END IF
         END DO ! i = 1, nlines
      END IF ! (debug)

      imesh = 0              ! 1st mesh grid ---------------------------
      ya(:,imesh) = ya0 + MATMUL(grY0(:,1:dimp,imesh),s(1:dimp))
      IF (nmesh>1) THEN      ! further mesh grids ----------------------
         DO imesh = 1,nmesh-1
            ya(:,imesh) = MATMUL(grY0(:,:,imesh), &
                               s(dimp+(imesh-1)*neqn+1:dimp+imesh*neqn))
            ! ya(:,imesh) = grY0(:,:,imesh)^{-1} * s
         END DO ! imesh = 1,nmesh-1
      END IF ! (nmesh>1)

      IF (verbose) THEN
        DO imesh = 0, nmesh-1
          WRITE(*,'(A,F8.6,A)') ' estimation at ',mesh(imesh), &
                                   ' is                   (should be)'
          DO i = 1, neqn
            WRITE(*,*) ya(i,imesh), '                   (', &
                                                         y0(i,imesh),')'
          END DO
          WRITE(*,*) 'with global error ', &
                                    MAXVAL(ABS(ya(:,imesh)-y0(:,imesh)))
        END DO ! imesh = 0, nmesh-1
      END IF ! (verbose)

      IF (MAXVAL(ABS(y0(:,0:nmesh-1))) <= EPSILON(1.0_dp) ) THEN
         WRITE(*,*) 'absolute error in all initial values:', &
                            MAXVAL(ABS(y0(:,0:nmesh-1)-ya(:,0:nmesh-1)))
      ELSE
         amaxerr = 0.0_dp
         DO imesh = 0, nmesh-1
            maxerr = MAXVAL(ABS(y0(:,imesh)-ya(:,imesh)))/&
                                                MAXVAL(ABS(y0(:,imesh)))
            IF (maxerr > amaxerr) amaxerr = maxerr
         END DO ! imesh = 0, nmesh-1
         WRITE(*,*) 'relative error in all initial values:', amaxerr
      END IF

      ! do damped Newton's method to find solution of non linear eqs
      lambda = dumping_factor_newton/2.0_dp ! initialise damping factor

      ! calculate the 1st Newton-residual F(s)
      imesh = 0              ! 1st mesh grid ---------------------------
      CALL big_init(dimp)    ! make dimensions available to other module
      yb = ya(:,imesh)       ! get IVs
      tol = basictol         ! adjust global error tolerance
      ifail = 0
      CALL ivp_ieuleridec(box_fcn, mesh(imesh), yb, mesh(imesh+1), &
           h(imesh), hmin, tol, eps, box_linear, maxidec, pord, &
           ifail, box_dfcn, dummy1, dummy2)
      IF (ifail /= 0) CALL error(ifail,.true.)
      IF (verbose) WRITE(*,*) 'used stepsize: ', h(imesh)/pord
      IF (nmesh == 1) THEN   ! just single shooting as special case
         F = MATMUL(Ba(dimq+1:neqn,:),ya(:,imesh)) + &
                        MATMUL(Bb(dimq+1:neqn,:),yb) - beta(dimq+1:neqn)
      ELSE                  
         F(1:neqn) = -yb     ! first lines
      END IF ! (nmesh == 1)

      IF (nmesh > 1) THEN    ! further mesh grids ----------------------
        CALL big_init(neqn)  ! make dimensions available to other module
        DO imesh = 1, nmesh-1
          j = (imesh-1)*neqn
          k = imesh*neqn
          l = (imesh+1)*neqn
          yb = ya(:,imesh)   ! get IVs
          tol = basictol     ! adjust global error tolerance
          ifail = 0
          CALL ivp_ieuleridec(box_fcn, mesh(imesh), yb, mesh(imesh+1), &
               h(imesh), hmin, tol, eps, box_linear, maxidec, pord, &
               ifail, box_dfcn, dummy1, dummy2)
          IF (ifail /= 0) CALL error(ifail,.true.)
          IF (verbose) WRITE(*,*) 'used stepsize: ', h(imesh)/pord
          F(j+1:k) = F(j+1:k) + ya(:,imesh)
          IF (imesh == nmesh-1) THEN ! last line
            F(k+1:k+dimp) = MATMUL(Ba(dimq+1:neqn,:),ya(:,0)) + &
                        MATMUL(Bb(dimq+1:neqn,:),yb) - beta(dimq+1:neqn)
          ELSE
            F(k+1:l) = -yb
          END IF ! last line
        END DO ! imesh = 1, nmesh-1
      END IF ! (nmesh > 1)
      ya(:,nmesh) = yb       ! store last value
      nrmF = MAXVAL(ABS(F))  ! norm of residual
      IF (verbose) WRITE(*,*) '1st Newton Residual |F(s)|', nrmF

!     IF (initloop>0) name1(4:4) = char(k+64)
      IF (initloop == 0) OPEN(14, FILE = name1) ! [iter, |res|, |incr|]
!     IF (initloop>0) name4(4:4) = char(k+64)
      IF (initloop == 0) OPEN(17, FILE = name4) ! [|res|, p, c]

      iter = 0
      DO                     ! start Newton Iteration
        iter = iter + 1
        WRITE(*,*) iter,'. Newton iteration'

        ! solve for F'(s) using discrete differential quotients
        deltaF = 0.0_dp

        imesh = 0            ! 1st mesh grid ---------------------------
        CALL big_init(dimp)
        ALLOCATE(Ze(dimp*neqn), Za(dimp*neqn), delta(dimp))
        DO i = 1, dimp
          delta(i) = MAX(1.0_dp,ABS(s(i))) * sqrteps
          Za((i-1)*neqn+1:i*neqn)=ya(:,imesh) + delta(i)*grY0(:,i,imesh)
        END DO
        Ze = Za
        tol = basictol       ! adjust global error tolerance
        ifail = 0
        CALL ivp_ieuleridec(bign_fcn, mesh(imesh), Ze, mesh(imesh+1), &
             h2(imesh), hmin, tol, eps, box_linear, maxidec, pord, &
             ifail, bign_dfcn, dummy1, dummy2)
        IF (ifail /= 0) CALL error(ifail,.true.)
        IF (verbose) WRITE(*,*) 'used stepsize: ', h2(imesh)/pord

        ! store values in deltaF for LES deltaF*deltas=-F
        IF (nmesh == 1) THEN ! just single shooting as special case
          DO i = 1, dimp
            yb(1:dimp) = - beta(dimq+1:neqn) + &
                    MATMUL(Ba(dimq+1:neqn,:), Za((i-1)*neqn+1:i*neqn)) &
                    + MATMUL(Bb(dimq+1:neqn,:), Ze((i-1)*neqn+1:i*neqn))
            deltaF(:,i) = (yb(1:dimp) - F)/delta(i)
!           WRITE(*,*) 'DeltaF(:,i)', deltaF(:,i)
          END DO ! quotients
        ELSE                 ! first and last submatrix
          DO i = 1, dimp
            yb = Ze((i-1)*neqn+1:i*neqn) + F(1:neqn) - ya(:,imesh+1)
            deltaF(1:neqn,i) = -yb/delta(i)
!           WRITE(*,*) 'DeltaF(:,i)', deltaF(1:neqn,i)
          END DO ! quotients
          deltaF((nmesh-1)*neqn+1:(nmesh-1)*neqn+dimp,1:dimp) = &
                          MATMUL(Ba(dimq+1:neqn,:),grY0(:,1:dimp,imesh))
        END IF ! (nmesh == 1)
        DEALLOCATE(delta, Za, Ze)

        IF (nmesh > 1) THEN  ! further mesh grids ----------------------
          CALL big_init(neqn)
          ALLOCATE(Ze(neqn*neqn), Za(neqn*neqn), delta(neqn))
          DO imesh = 1, nmesh-1
            j = (imesh-1)*neqn
            k = imesh*neqn
            l = (imesh+1)*neqn
            ! store IVs grY0 in deltaF for LES deltaF*deltas=-F
            deltaF(j+1:k,dimp+j+1:dimp+k) = grY0(:,:,imesh)
            DO i = 1, neqn
              delta(i) = MAX(1.0_dp,ABS(s(dimp+j+i))) * sqrteps
              Za((i-1)*neqn+1:i*neqn) = ya(:,imesh) + &
                                                delta(i)*grY0(:,i,imesh)
            END DO
            Ze = Za
            tol = basictol   ! adjust global error tolerance
            ifail = 0
            CALL ivp_ieuleridec(bign_fcn,mesh(imesh),Ze,mesh(imesh+1), &
                 h2(imesh), hmin, tol, eps, box_linear, maxidec, pord, &
                 ifail, bign_dfcn, dummy1, dummy2)
            IF (ifail /= 0) CALL error(ifail,.true.)
            IF (verbose) WRITE(*,*) 'used stepsize: ', h2(imesh)/pord

            ! store values in deltaF for LES deltaF*deltas=-F
            IF (imesh==nmesh-1) THEN ! last loop, use BVPs
              DO i = 1, neqn
                yb = (Ze((i-1)*neqn+1:i*neqn) - ya(:,nmesh))/delta(i)
                deltaF(k+1:k+dimp,dimp+j+i)=MATMUL(Bb(dimq+1:neqn,:),yb)
!               WRITE(*,*) 'DeltaF(:,i)', deltaF(k+1:k+dimp,dimp+j+i)
              END DO ! quotients
            ELSE
              DO i = 1, neqn
                yb = Ze((i-1)*neqn+1:i*neqn) + F(k+1:l) - ya(:,imesh+1)
                deltaF(k+1:l,dimp+j+i) = -yb/delta(i)
!               WRITE(*,*) 'DeltaF(:,i)', deltaF(k+1:l,dimp+j+i)
              END DO ! quotients
            END IF ! (imesh==nmesh-1)
          END DO ! imesh=1,nmesh
          DEALLOCATE(delta, Za, Ze)
        END IF ! (nmesh > 1)

        ! solve LES deltaF*deltas=-F to calculate increment deltas
        deltas = -F
        WRITE(*,*) "L1-norm F'(s): ",m1abs(deltaF)
        WRITE(*,*) "L1-norm inverse(F'(s)): ",m1absinv(deltaF)
        WRITE(*,*) "L1-condition F'(s): ",L1cond(deltaF)
        ALLOCATE(luindx(nlines))
        CALL ludcmp(deltaF,luindx)
        CALL lubksb(deltaF,luindx,deltas)
        nrmds = MAXVAL(ABS(deltas)) ! norm of Newton increment

        ! if we have enough values, calculate order of Newton's method
        IF ((initloop == 0).AND.(iter > 2).AND.(nrmds/=0.0_dp).AND.&
           (nrmo1ds/=0.0_dp)) THEN
           IF ((nrmo2ds/nrmo1ds-1.0_dp) /= 0.0_dp) THEN
              dsp = log(nrmo1ds/nrmds)/log(nrmo2ds/nrmo1ds)
              dsc = nrmds/(exp(log(nrmo1ds)*dsp))
           ELSE
              dsp = 0.0_dp
              dsc = 0.0_dp
           END IF
        ELSE
           dsp = 0.0_dp
           dsc = 0.0_dp
        END IF
        nrmo2ds = nrmo1ds
        nrmo1ds = nrmds      ! store old increment length

        ! exit iteration if accurate
        WRITE(*,*) 'Newton residual: ',nrmF
        WRITE(*,*) 'Newton increment:',nrmds
        IF (initloop == 0) WRITE(14, *) iter, nrmF, nrmds
        IF (initloop == 0) WRITE(17, *) nrmF, dsp, dsc ! [|res|, p, c]

        IF ((nrmF<Beps).and.(nrmds<Beps*(1.0_dp+nrmF))) THEN
            DEALLOCATE(luindx)
            EXIT
        END IF

        ! update solution and iterate damping factors
        lambda = MIN(1.0_dp,lambda*2.0_dp)
        lindex = 0
        ALLOCATE(slambda(nlines),deltaslam(nlines))
        DO
         IF (lindex>0) WRITE(*,*) lindex,'. damped Newton iteration'
         lindex = lindex + 1
         slambda = s + deltas*lambda

         imesh = 0          ! 1st mesh grid ---------------------------
         ya(:,imesh) = ya0+MATMUL(grY0(:,1:dimp,imesh),slambda(1:dimp))
         IF (nmesh>1) THEN  ! further mesh grids ----------------------
           DO imesh = 1,nmesh-1
             ya(:,imesh) = MATMUL(grY0(:,:,imesh), &
                        slambda(dimp+(imesh-1)*neqn+1:dimp+imesh*neqn))
           END DO ! imesh = 1,nmesh-1
         END IF ! (nmesh>1)

         ! calculate new Newton-residual
         imesh = 0           ! 1st mesh grid ---------------------------
         CALL big_init(dimp)
         yb = ya(:,imesh)    ! get IVs
         tol = basictol      ! adjust global error tolerance
         ifail = 0
         CALL ivp_ieuleridec(box_fcn, mesh(imesh), yb, mesh(imesh+1), &
              h(imesh), hmin, tol, eps, box_linear, maxidec, pord, &
              ifail, box_dfcn, dummy1, dummy2)
         IF (ifail /= 0) THEN
           CALL error(ifail,.false.)
         ELSE
           IF (verbose) WRITE(*,*) 'used stepsize: ', h(imesh)/pord
           IF (nmesh == 1) THEN ! just single shooting as special case
             F = MATMUL(Ba(dimq+1:neqn,:),ya(:,imesh)) + &
                        MATMUL(Bb(dimq+1:neqn,:),yb) - beta(dimq+1:neqn)
           ELSE
             F(1:neqn) = -yb  ! first lines
           END IF ! (nmesh == 1)
         END IF ! (ifail /= 0, 1st mesh grid)

         IF ((ifail == 0).AND.(nmesh > 1)) THEN ! further mesh grids ---
           CALL big_init(neqn)
           DO imesh = 1, nmesh-1
             j = (imesh-1)*neqn
             k = imesh*neqn
             l = (imesh+1)*neqn
             yb = ya(:,imesh) ! get IVs
             tol = basictol  ! adjust global error tolerance
             ifail = 0
             CALL ivp_ieuleridec(box_fcn, mesh(imesh),yb,mesh(imesh+1),&
                  h(imesh), hmin, tol, eps, box_linear, maxidec, pord, &
                  ifail, box_dfcn, dummy1, dummy2)
             IF (ifail /= 0) THEN
               CALL error(ifail,.false.)
               EXIT
             ELSE
               IF (verbose) WRITE(*,*) 'used stepsize: ', h(imesh)/pord
               F(j+1:k) = F(j+1:k) + ya(:,imesh)
               IF (imesh == nmesh-1) THEN ! last line
                 F(k+1:k+dimp) = MATMUL(Ba(dimq+1:neqn,:),ya(:,0)) + &
                        MATMUL(Bb(dimq+1:neqn,:),yb) - beta(dimq+1:neqn)
               ELSE
                 F(k+1:l) = -yb
               END IF ! last line
             END IF ! (ifail /= 0)
           END DO ! imesh = 1, nmesh-1
         END IF ! (nmesh > 1)

         IF (ifail == 0) THEN
           ya(:,nmesh) = yb  ! store last value

           ! check "monotony" of residuals by simplified increments
           deltaslam = -F    ! we need -F to calulate  increment
           CALL lubksb(deltaF,luindx,deltaslam)
           IF (verbose) WRITE(*,*) MAXVAL(ABS(deltaslam)), ' <= ', &
                                            (1.0_dp-lambda/2.0_dp)*nrmds
           IF(MAXVAL(ABS(deltaslam))<=(1.0_dp-lambda/2.0_dp)*nrmds) EXIT
           ! "monotony test" is true therefore take this solution
         END IF ! (ifail == 0)

         ! shorten increment's length by 2
         lambda = lambda/2.0_dp
         IF (lindex > max_lambda_steps) THEN
            ifail = 6        ! error, residual did not decrease
            CALL error(ifail,.true.)
         END IF
        END DO ! lambda dumping
        s = slambda
        DEALLOCATE(deltaslam,slambda,luindx)
        nrmF = MAXVAL(ABS(F)) ! norm of residual

        IF (verbose) THEN
          DO imesh = 0, nmesh
            WRITE(*,'(A,F8.6,A)') ' solution at ',mesh(imesh),&
                                   ' is                     (should be)'
            DO i = 1, neqn
              WRITE(*,*) ya(i,imesh), '                   (', &
                                                         y0(i,imesh),')'
            END DO
            WRITE(*,*) 'with global error ', &
                                    MAXVAL(ABS(ya(:,imesh)-y0(:,imesh)))
          END DO ! imesh = 0, nmesh
        END IF ! (verbose)

        IF (iter > max_newton_steps) THEN
           ifail = 5         ! error, too many iterations
           CALL error(ifail,.true.)
        END IF
      END DO ! Do Newton iterations
      IF (initloop == 0) CLOSE(14)
      IF (initloop == 0) CLOSE(17)

      DEALLOCATE(deltaF, deltas, F, s, h2, grY0)
      WRITE(*,*) 'number of Newton iterations:', iter
   END IF ! (box_linear)

! ----------------------------------------------------------------------
!        repeat integration to calculate complete solution curve
! ----------------------------------------------------------------------
   WRITE(*,*) 'calculating complete solution on [a,b]'

   amaxerr = 0.0_dp
   amaxerrx = a

   imesh = 0                 ! 1st mesh grid ---------------------------
   WRITE(*,'(A,F8.6,A)') ' solution at ',mesh(0),&
                                   ' is                     (should be)'
   DO i = 1, neqn
      WRITE(*,*) ya(i,imesh), '                   (', y0(i,imesh),')'
   END DO
   WRITE(*,*) 'with global error ', MAXVAL(ABS(ya(:,imesh)-y0(:,imesh)))

   DO imesh = 1, nmesh
!    IF (initloop == 0) log_file = .true. ! enable plots
     yb = ya(:,imesh-1)      ! set IVs
     tol = basictol          ! adjust global error tolerance
     ifail = 0
     CALL ivp_ieuleridec(box_fcn, mesh(imesh-1), yb, mesh(imesh), &
          h(imesh-1), hmin, tol, eps, box_linear, maxidec, pord, ifail,&
          box_dfcn, box_true, box_truelocal, maxerr=maxerr, &
          maxerrx=maxerrx)
     IF (ifail /= 0) CALL error(ifail,.true.)
     IF (verbose) WRITE(*,*) 'used stepsize: ', h(imesh-1)/pord
     IF (maxerr > amaxerr) THEN
        amaxerr = maxerr
        amaxerrx = maxerrx
     END IF ! (maxerr > amaxerr)
     WRITE(*,'(A,F8.6,A)') ' solution at ',mesh(imesh),&
                                   ' is                     (should be)'
     IF ((nmesh < 8 ).AND.(initloop==0)) THEN
        DO i = 1, neqn
           WRITE(*,*) yb(i), '                   (', y0(i,imesh),')'
        END DO
     END IF
     WRITE(*,*) 'with difference to first approximation ', &
                                             MAXVAL(ABS(ya(:,imesh)-yb))
     WRITE(*,*) 'with global error ', MAXVAL(ABS(yb-y0(:,imesh)))
     WRITE(*,*) 'and its maximum ', maxerr, ' at ', maxerrx
     IF (.not. debug) WRITE(*,*) 'with global error estimation', tol
   END DO ! imesh = 0, nmesh-1

   IF (initloop > 0) THEN
      DEALLOCATE(ya)
      WRITE(*,*) 'the maximum of the global error is ', amaxerr
      WRITE(15, *) maxinitial, iter, amaxerr
      IF (iloop == 0 ) THEN
         maxinitial = maxinitstart
      ELSE
         maxinitial = maxinitial * maxinitmul
      END IF
      IF (iloop < initloop) THEN
         WRITE(*,*) 'relative error of estimated initial values:'
         WRITE(*,*) maxinitial
      END IF
   END IF
   END DO ! iloop = 0,initloop
   IF (initloop > 0) CLOSE(15)

   WRITE(*,*) 'the maximum of the global error is ', amaxerr
   WRITE(*,*) 'and it is at the position ' ! ,amaxerrx
   WRITE(*,*) amaxerrx

   WRITE(*,*) 'FUER ORDALL'
!  order of error |y(a)| and |y(b)|:
!  WRITE(*,*) MAXVAL(ABS(ya(:,0)-y0(:,0))), &
!             MAXVAL(ABS(ya(:,nmesh)-y0(:,nmesh)))
!  order of error |y(b)| and |y([a,b])|:
!  WRITE(*,*) MAXVAL(ABS(ya(:,nmesh)-y0(:,nmesh))), amaxerr
!  order of error |y(a)| and |y([a,b])|:
!  WRITE(*,*) MAXVAL(ABS(ya(:,0)-y0(:,0))), amaxerr
!  order of error |y(t_i)| for all t_i and |y([a,b])|:
   WRITE(*,*) MAXVAL(ABS(ya(:,0:nmesh)-y0(:,0:nmesh))), amaxerr

   DEALLOCATE(ya, h, y0, mesh)

   STOP

   CONTAINS
! ----------------------------------------------------------------------
SUBROUTINE write_mat(A)
! print a matrix for debugging purposes
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: A(:,:)
   INTEGER(KIND = i4b) :: i
   DO i = 1, SIZE(A,1)
      IF (SIZE(A,1)>6) THEN
         WRITE(*,*) MAXVAL(ABS(A(i,:)))
      ELSE
         WRITE(*,*) A(i,:)
      END IF
   END DO
   RETURN
END SUBROUTINE write_mat
! ----------------------------------------------------------------------
SUBROUTINE error(ifail, bstop)
! write out error message and stop program
   IMPLICIT NONE
   INTEGER(KIND = i4b), INTENT(in) :: ifail
!        ifail ... error variable, set due to several errors
   LOGICAL, INTENT(in) :: bstop
!        bstop ... stop program
   IF ((ifail > -9).AND.(ifail < 5)) THEN
      WRITE(*,'(A,I1,A)') ' ERROR ',ifail,' in black-box:'
   ELSE
      WRITE(*,'(A,I1,A)') ' ERROR ',ifail,' in shooting procedure:'
   END IF
   SELECT CASE (ifail)
   CASE (0)
      WRITE(*,*) 'everything is OK (no error)'
   CASE (-1)
      WRITE(*,*) 'an illegal interval range was specified (b <= a)'
   CASE (-2)
      WRITE(*,*) 'an illegal smallest step size was specified'
   CASE (-3)
      WRITE(*,*) 'an illegal starting step size was specified'
   CASE (-4)
      WRITE(*,*) 'an illegal global tolerance was specified (tol <= 0)'
   CASE (-5)
      WRITE(*,*) 'an illegal Newton-tolerance was specified'
      WRITE(*,*) '(Newton-tol >= global tol or Newton-tol <= 0)'
   CASE (-6)
     WRITE(*,*) 'an illegal order of method or polynomial was specified'
     WRITE(*,*) '(order of polynomials <= order of method)'
   CASE (-7)
      WRITE(*,*) 'an illegal maximum step size was specified'
   CASE (-8)
     WRITE(*,*) 'an illegal switch point was specified (x not in (a,b))'
   CASE (1)
      WRITE(*,*) 'IVP solver could not obtain specified accuracy', &
                 ' (h = h(global tol) < h_min)'
   CASE (2)
      WRITE(*,*) 'the Newton-iteration needed too many steps', &
                 ' (Newton did not converge)'
   CASE (3)
    WRITE(*,*) 'the global error did not decrease during IDeC iteration'
    WRITE(*,*) '(IDeC did not converge for h to 0)'
   CASE (4)
    WRITE(*,*) 'the Newton-residuals did not decrease during iterations'
    WRITE(*,*) '(Newton-residuals were not decreasing in monotonic way)'
   CASE (5)
      WRITE(*,*) 'the Newton-iteration needed too many steps', &
                 ' (Newton did not converge)'
   CASE (6)
    WRITE(*,*) 'the Newton-residuals did not decrease during iterations'
    WRITE(*,*) '(Newton-residuals were not decreasing in monotonic way)'
   END SELECT
   IF ( (bstop) .AND. (ifail /= 0) ) THEN
      WRITE(*,*) 'FUER ORDALL (dummy)'
      WRITE(*,*) 0.0_dp, 0.0_dp
      STOP
   END IF
   RETURN
END SUBROUTINE error

END PROGRAM MSHT
